"""
The productstatus Python module is an object-based HTTP REST interface to the
Productstatus service. Use the `Api` class in the `productstatus.api` module.
"""

def datainstance_has_complete_file_count(datainstance):
    service_backend_uri = datainstance.servicebackend.resource_uri
    format_uri = datainstance.format.resource_uri
    return datainstance.data.productinstance.complete[service_backend_uri][format_uri]['file_count']



import datetime
import dateutil.tz


def build_url(*args):
    """
    Join a list of strings into a slash-separated string, stripping any leading
    slashes but including the trailing slash.
    """
    return '/'.join([x.strip('/') for x in args]) + '/'


def get_utc_now():
    """
    Return a time-zone aware DateTime object with the current date and time
    """
    return datetime.datetime.utcnow().replace(tzinfo=dateutil.tz.tzutc())


class SerializeBase(object):
    __serializable__ = []

    def serialize(self):
        """
        Create JSON encodable representation of internal data structure.
        """
        serialized = {}
        for key in self.__serializable__:
            func_name = 'serialize_' + key
            func = getattr(self, func_name, None)
            serialized[key] = getattr(self, key, None)
            if callable(func):
                serialized[key] = func(serialized[key])
            elif hasattr(serialized[key], 'serialize'):
                serialized[key] = serialized[key].serialize()
        return serialized

    def unserialize(self, data):
        """
        Load internal data structure from JSON decoded dictionary.
        """
        for key in self.__serializable__:
            func_name = 'unserialize_' + key
            func = getattr(self, func_name, None)
            if callable(func):
                value = func(data[key])
            else:
                value = data[key]
            setattr(self, key, value)

    def _serialize_datetime(self, value):
        """
        Return a time zone-aware ISO 8601 string.
        """
        utc_time = value.astimezone(tz=dateutil.tz.tzutc())
        return utc_time.isoformat().replace(' ', 'T').replace('+00:00', 'Z')

    def _unserialize_datetime(self, value):
        """
        Return a DateTime object from a ISO 8601 string.
        """
        return dateutil.parser.parse(value)



"""
Command-line interface to the Productstatus server.
"""

import sys
import traceback
import json
import argparse

import productstatus.api
import productstatus.exceptions


class Client(object):
    def __init__(self):
        self.subparsers = {}

    def setup_basic_parser(self):
        self.parser = argparse.ArgumentParser(add_help=False)
        self.parser.add_argument('server', help='Productstatus server to communicate with')
        self.parser.add_argument('--help', required=False, action='store_true', help='Print help')
        self.parser.add_argument('--username', required=False, help='Productstatus user name')
        self.parser.add_argument('--api_key', required=False, help='Productstatus API key')

    def setup_preliminary_parser(self):
        self.setup_basic_parser()
        self.parser.add_argument('...', nargs=argparse.REMAINDER)

    def setup_sub_commands(self):
        self.setup_preliminary_parser()
        args = self.parser.parse_args()
        self.api = productstatus.api.Api(args.server, username=args.username, api_key=args.api_key)
        self.api._get_schema_from_server()
        self.main_schema = self.api._schema

        # Create a new parser with all sub-commands
        self.setup_basic_parser()
        self.subparser = self.parser.add_subparsers(dest='_subparser')
        for key in self.main_schema.keys():
            self.subparsers[key] = self.subparser.add_parser(
                key,
                add_help=False,
                help='Interact with the %s resource' % key,
                )

    def pprint_json_string(self, json_data):
        hash_ = json.loads(json_data)
        self.pprint(hash_)

    def pprint(self, hash_):
        print json.dumps(hash_, sort_keys=True, indent=4, separators=(',', ': '))

    def args_in_schema(self, args_dict):
        schema_keys = self.schema['fields'].keys()
        a = {}
        for key in args_dict:
            if args_dict[key] is None:
                continue
            # Remote queries may have modifier such as field__gte, whereas the
            # schema only contains the 'field' part.
            base_key = key.split('__')[0]
            if base_key in schema_keys:
                a[key] = args_dict[key]
        return a

    def exec_get(self, args_dict):
        resource = self.collection[args_dict['uuid']]
        serialized = resource._serialize()
        self.pprint_json_string(serialized)

    def exec_search(self, args_dict):
        args = self.args_in_schema(args_dict)
        qs = self.collection.objects
        qs.filter(**args)
        resources = [json.loads(x._serialize()) for x in qs]
        self.pprint(resources)

    def exec_create(self, args_dict):
        args = self.args_in_schema(args_dict)
        resource = self.collection.create()
        [setattr(resource, key, value) for key, value in args.iteritems()]
        resource._unserialize()
        resource.save()
        serialized = resource._serialize()
        self.pprint_json_string(serialized)

    def setup_parameters(self, subcommand):
        self.setup_basic_parser()
        self.parser.add_argument(subcommand, help='The literal string "%s"' % subcommand)
        self.collection = getattr(self.api, subcommand)
        self.schema = self.collection.schema
        self.subparser = self.parser.add_subparsers(dest='action')
        self.subparsers = {}
        for action in ['search', 'create', 'get']:
            self.subparsers[action] = self.subparser.add_parser(action)
            self.subparsers[action].set_defaults(func=getattr(self, 'exec_%s' % action))
        for action in ['search', 'create']:
            for key, value in self.schema['fields'].iteritems():
                self.subparsers[action].add_argument('--%s' % key, help=value['help_text'])
        for action in ['get']:
            self.subparsers[action].add_argument('uuid', help='The UUID of a %s object' % subcommand)

    def _exec(self, args, args_dict):
        """
        Execute command set by func attribute.
        Catch exceptions, print traceback and exit program with proper exit_code.
        """
        try:
            args.func(args_dict)
        except productstatus.exceptions.UnauthorizedException:
            self._exit(2)
        except productstatus.exceptions.NotFoundException:
            self._exit(3)
        except productstatus.exceptions.ServiceUnavailableException:
            self._exit(4)
        except productstatus.exceptions.ProductstatusException:
            self._exit(255)
        except:
            self._exit(1)

    def _exit(self, exit_code):
        traceback.print_exc()
        sys.exit(exit_code)

    def main(self):
        self.setup_sub_commands()
        args, unknown = self.parser.parse_known_args()
        self.setup_parameters(args._subparser)
        args, unknown = self.parser.parse_known_args()
        args_dict = vars(args)
        # This very filthy filter ensures that remote filter extensions are
        # applied magically, at the cost of breaking local syntax checking.
        for i in xrange(0, len(unknown), 2):
            args_dict[unknown[i].strip('--')] = unknown[i+1]
        if args.help:
            self.parser.print_help()
            return

        self._exec(args, args_dict)

if __name__ == '__main__':
    client = Client()
    client.main()



"""
This module contains exception objects used in the Productstatus client library.
"""


class ProductstatusException(Exception):
    """
    Thrown when there is an error relating to getting data from the REST API.
    """
    pass


class InvalidFilterDataException(ProductstatusException):
    """
    Thrown when filter data for query sets are invalid.
    """
    pass


class InvalidFilterDataException(ProductstatusException):
    """
    Thrown when filter data for query sets are invalid.
    """
    pass


class ClientErrorException(ProductstatusException):
    """
    Thrown when the server returns a 4xx error.
    """
    pass


class UnauthorizedException(ClientErrorException):
    """
    Thrown when the server returns 401.
    """
    pass


class NotFoundException(ClientErrorException):
    """
    Thrown when the server returns 404.
    """
    pass


class ResourceTypeNotFoundException(NotFoundException):
    """
    Thrown when a resource type can not be found on the server.
    """
    pass


class ResourceNotFoundException(NotFoundException):
    """
    Thrown when a requested resource is not available on the server.
    """
    pass


class ServiceUnavailableException(ProductstatusException):
    """
    Thrown when the server returns a 5xx error.
    """
    pass


class UnserializeException(ProductstatusException):
    """
    Thrown when the data from the REST API could not be decoded.
    """
    pass


class InvalidResourceException(ProductstatusException):
    """
    Thrown when the server returns an invalid resource.
    """
    # FIXME: DEPRECATED
    pass


class EventException(ProductstatusException):
    """
    Parent class for all exceptions related to Productstatus events.
    """
    pass


class EventTimeoutException(EventException):
    """!
    @brief Thrown when an event is not available on the Kafka socket.
    """
    pass



"""!
The productstatus.event module is an interface to the Kafka distributed commit log
where Productstatus server publishes its events
"""

import kafka
import json
import uuid

import productstatus.exceptions


def unserialize(message):
    return json.loads(message.decode('utf-8'))


class Message(dict):
    """!
    @brief A Kafka event message.
    """

    def __getattr__(self, name):
        if name in self:
            return self[name]
        raise AttributeError('Attribute %s does not exist' % name)


class Listener(object):
    """!
    @brief Kafka event listener client that provides a simple interface for
    fetching the next event from the queue.
    """

    def __init__(self, *args, **kwargs):
        """!
        @brief Set up a connection to the Kafka instance on Productstatus server.

        Takes the same parameters as the KafkaConsumer() constructor. Client
        and group UUIDs will be auto-generated if not specified.
        """

        if 'client_id' not in kwargs or not kwargs['client_id']:
            kwargs['client_id'] = unicode(uuid.uuid4())

        if 'group_id' not in kwargs or not kwargs['group_id']:
            kwargs['group_id'] = unicode(uuid.uuid4())

        kwargs['enable_auto_commit'] = False
        kwargs['value_deserializer'] = unserialize

        self.client_id = kwargs['client_id']
        self.group_id = kwargs['group_id']

        self.json_consumer = kafka.KafkaConsumer(*args, **kwargs)

    def get_next_event(self):
        """!
        @brief Block until a message is received, or a timeout is reached, and
        return the message object. Raises an exception if a timeout is reached.
        @returns a Message object.
        """
        try:
            return Message(self.json_consumer.next().value)
        except StopIteration:
            raise productstatus.exceptions.EventTimeoutException('Timeout while waiting for next event')

    def save_position(self):
        """!
        @brief Store the client's position in the message queue.

        When this function is used, Kafka will store the client's message queue
        position. Thus, next time the client is run, it will resume from the
        next message. To use this function properly, you must set `client_id`
        and `group_id` when instantiating the Listener object.
        """
        self.json_consumer.commit()



import copy
import requests
import requests.auth
import json
import logging
import datetime
import dateutil.parser
import dateutil.tz

import productstatus.utils
import productstatus.exceptions
import productstatus.event


SERVICE_UNAVAILABLE_EXCEPTIONS = (requests.exceptions.RequestException,
                                  requests.exceptions.Timeout,
                                  )


class Api(object):
    """
    This class provides fluent access to the Productstatus REST API. Resource
    collections are exposed as members of the Api class, and specific resources
    can be retrieved by indexes.

    The API client tries to be as lazy as possible, and does not make any
    unneccessary requests to the server.

    Examples of use:

    api = Api('http://localhost:8000')
    products = api.product
    arome = products['66340f0b-2c2c-436d-a077-3d939f4f7283']
    print arome.grid_resolution
    """

    def __init__(self, base_url, verify_ssl=True, username=None, api_key=None, timeout=3):
        """
        Initialize the Api class.

        @param base_url The root URL where the Productstatus server serves data.
        @param verify_ssl Whether or not to verify SSL certificates.
        @param username Client API username.
        @param api_key Client API key.
        @param timeout Request timeout in seconds.
        """
        self._base_url = base_url
        self._url_prefix = '/api/v1/'
        self._url = productstatus.utils.build_url(self._base_url, self._url_prefix)
        self._verify_ssl = verify_ssl
        self._timeout = timeout
        self._session = requests.Session()
        self._session.verify = self._verify_ssl
        self._session.headers.update({'content-type': 'application/json'})
        self._event_listener = None
        if username and api_key:
            self._session.auth = TastypieApiKeyAuth(username, api_key)
        self._resource_collection = {}
        self._schema = {}

    def get_event_listener_configuration(self):
        """!
        @brief Fetch the message queue configuration from the Productstatus server.
        """
        return self.kafka['default']

    def get_event_listener(self, **kwargs):
        """!
        @brief Instantiate a message delivery object of type with configuration
        retrieved from the Productstatus server.
        @returns A productstatus.event.Listener object.
        """
        if not self._event_listener:
            configuration = self.get_event_listener_configuration()
            kwargs['bootstrap_servers'] = configuration.brokers
            self._event_listener = productstatus.event.Listener(str(configuration.topic), **kwargs)
        return self._event_listener

    def _do_request(self, method, *args, **kwargs):
        """
        Run a request through the requests API. This function wraps
        self.session.{get,post,patch,put,delete} and adds exception handling.

        Returns a response object.
        """
        if 'timeout' not in kwargs:
            kwargs['timeout'] = self._timeout
        try:
            response = self._session.request(method, *args, **kwargs)
        except SERVICE_UNAVAILABLE_EXCEPTIONS, e:
            raise productstatus.exceptions.ServiceUnavailableException(
                "Could not perform request: %s" % unicode(e)
            )

        self._raise_response_exceptions(response)
        return response

    def _get_response_data(self, response):
        """
        Get unserialized contents from a response object.
        """
        if response.content:
            return self._unserialize(response.content)
        return response.content

    def _raise_response_exceptions(self, response):
        """
        Conditionally raise exceptions if a response object has errors.
        """
        if response.status_code < 400:
            return
        if response.status_code >= 500:
            exception = productstatus.exceptions.ServiceUnavailableException
        elif response.status_code == 401:
            exception = productstatus.exceptions.UnauthorizedException
        elif response.status_code == 404:
            exception = productstatus.exceptions.NotFoundException
        else:
            exception = productstatus.exceptions.ClientErrorException
        raise exception(response.text)

    def _unserialize(self, data):
        """
        Convert JSON encoded data into a dictionary.
        """
        try:
            return json.loads(data)
        except ValueError, e:
            raise productstatus.exceptions.UnserializeException(e)

    def _get_schema_from_server(self):
        """
        Retrieve a list of possible resource types from the server.
        """
        response = self._do_request('get', self._url)
        self._schema = self._get_response_data(response)

    def _validate_url_component(self, name):
        """
        Raise an exception if an URL slug cannot be used to determine the
        location of a resource or collection.
        """
        if len(name) == 0:
            raise NameError('URL components cannot be empty')
        if name[0] == '_':
            raise NameError('URL components cannot begin with an underscore')

    def __getattr__(self, name):
        """
        Return a ResourceCollection object which can be used to retrieve
        resources. E.g. api.productinstance will point to /api/v1/productinstance/.
        """
        if name not in self._resource_collection.keys():
            if not self._schema:
                self._get_schema_from_server()
            if name not in self._schema.keys():
                # ignore Python internal functions
                if name[:2] == '__':
                    raise AttributeError('Attribute %s not found' % name)
                raise productstatus.exceptions.ResourceTypeNotFoundException(
                    "The resource '%s' is not supported by the Productstatus server" % name)
            self._resource_collection[name] = ResourceCollection(self, name)
        return self._resource_collection[name]

    def __getitem__(self, index):
        """
        Provide direct access to an API object. For instance, accessing
        api['/api/v1/product/66340f0b-2c2c-436d-a077-3d939f4f7283/'] will return a
        Resource object.
        """
        prefix_length = len(self._url_prefix)
        if self._url_prefix != index[:prefix_length]:
            raise KeyError('The URL %s is unsupported by this API instance; the URL must begin with %s' %
                           (index, self._url_prefix))
        components = index[prefix_length:].strip('/').split('/')

        # Sanity checks
        if len(components) == 0:
            return self
        for slug in components:
            self._validate_url_component(slug)

        # Load collection
        collection = getattr(self, components[0])
        if len(components) == 1:
            return collection

        # Load resource
        resource = collection[components[1]]
        if len(components) == 2:
            return resource

        raise NameError('The URL %s in unsupported by this API instance; the URL contains too many components' %
                        index)

    def __repr__(self):
        """
        Return a human-readable string representing this Api object.
        """
        return '<Productstatus API at %s>' % self._url


class QuerySet(object):
    """
    The QuerySet class facilitates listing and filtering a resource collection.

    Example usage:
    --------------
    qs = api.productinstance.objects  # instantiates a QuerySet
    qs.filter(product=api.product['66340f0b-2c2c-436d-a077-3d939f4f7283'], reference_time=datetime.datetime.now())
    qs.order_by('-version')  # order by version field, descending
    qs.limit(5)  # limit query to 5 results
    qs.count()  # returns total matches, regardless of limit
    qs[2]  # returns the 3rd element matching the criteria
    """

    def __init__(self, api, collection):
        self._api = api
        self._collection = collection
        self._filters = {}
        self._results = {}

    def _relative_item_index(self, index):
        """
        Convert an item's absolute list position to the relative list position
        in self._results['objects']. Returns None if the object is not locally cached.
        """
        if not self._results:
            return None
        offset = self._results['meta']['offset']
        limit = self._results['meta']['limit']
        if index < offset or index >= limit + offset:
            return None
        return index - offset

    def filter(self, **kwargs):
        """
        Add a search constraint and return a reference to self.
        """
        self._results = {}
        [self._add_filter(key, value) for key, value in kwargs.iteritems()]
        return self

    def all(self):
        """
        Deletes all search constraints and return a reference to self.
        """
        self._filters = {}
        return self.filter()

    def order_by(self, *args):
        """
        Apply list ordering. Sorted ascending by default. Prefix member names
        with a minus sign to have descending order.
        """
        return self.filter(order_by=list(args))

    def limit(self, limit):
        """
        Limit the number of results returned from the server.
        """
        return self.filter(limit=int(limit))

    def execute(self):
        """
        Fetch results from the server.
        """
        response = self._api._do_request('get', self._collection._url, params=self._filters)
        self._results = self._api._get_response_data(response)

    def execute_if_empty(self):
        """
        Ensure there exists some search results.
        """
        if not self._results:
            self.execute()

    def count(self):
        """
        Return the number of results in the search query.
        """
        self.execute_if_empty()
        return self._results['meta']['total_count']

    def _add_filter(self, key, value):
        """
        Add a filter to the search query, serializing if neccessary.
        """
        if isinstance(value, productstatus.api.Resource):
            self._filters[key] = value.id
        elif isinstance(value, datetime.datetime):
            if not value.tzname():
                raise productstatus.exceptions.InvalidFilterDataException(
                    'Cannot use a naive timestamp for filtering'
                )
            self._filters[key] = value.astimezone(dateutil.tz.tzutc()).strftime('%Y-%m-%dT%H:%M:%SZ')
        else:
            self._filters[key] = value

    def __getitem__(self, index):
        """
        Return the Resource of Nth index in the search results, running a
        remote request if needs be.
        """
        relative_index = self._relative_item_index(index)
        if relative_index is None or not self._results:
            self.filter(offset=index)
            self.execute()
            relative_index = self._relative_item_index(index)
        if relative_index is None:
            raise IndexError('Out of range: %d' % index)
        item = self._results['objects'][relative_index]
        return Resource(self._api, self._collection, item['id'], item)

    def __repr__(self):
        """
        Return a human-readable string representing this query set.
        """
        return '<QuerySet on %s>' % self._collection._url


class ResourceCollection(object):
    """
    The ResourceCollection class is used to retrieve resources from the REST
    API. Resources are accessed using indexes.
    For instance, resource['66340f0b-2c2c-436d-a077-3d939f4f7283'] will return
    a Resource object.
    """

    def __init__(self, api, resource_name):
        self._api = api
        self._resource_name = resource_name
        self._url = productstatus.utils.build_url(self._api._url, self._resource_name)
        self._schema_url = productstatus.utils.build_url(self._url, 'schema')
        self._schema = {}

    def create(self):
        """
        Create a new, temporary Resource object that might be saved, and thus
        stored on the server.
        """
        return Resource(self._api, self, None)

    def find_or_create(self, data, order_by=None):
        """
        Find or create a resource matching the given parameters in the
        `data` variable. If none is found, created and save one.
        @param data data to search for, or to set if nothing is found
        @param order_by ordering
        @returns a single Resource object.
        """
        # search for existing
        qs = self.objects
        qs.filter(**data)
        if order_by:
            qs.order_by(order_by)

        # create if not found
        if qs.count() == 0:
            logging.info('No matching %s resource found, creating...' % self._resource_name)
            resource = self.create()
            [setattr(resource, key, value) for key, value in data.iteritems()]
            resource.save()
            logging.info('%s: resource created' % resource)
        else:
            resource = qs[0]
            logging.info('%s: using existing resource' % resource)

        return resource

    def _get_schema_from_server(self):
        """
        Retrieve from the server the data model schema for this resource type.
        """
        response = self._api._do_request('get', self._schema_url)
        self._schema = self._api._get_response_data(response)

    def __getitem__(self, id):
        """
        Resource accessor. Will create and return a Resource instance pointing
        to a specific resource.
        """
        return Resource(self._api, self, id)

    def __getattr__(self, name):
        """
        Schema or query set accessor.

        Schema accessor returns a dictionary with the schema for this
        particular resource type. It is retrieved from the server unless it is
        locally cached.

        The query set accessor returns an object which is used to return a list
        of objects.
        """
        if name == 'schema':
            if not self._schema:
                self._get_schema_from_server()
            return self._schema
        elif name == 'objects':
            return QuerySet(self._api, self)

        raise AttributeError('Attribute does not exist: %s' % name)

    def __repr__(self):
        """
        Return a human-readable string representing this resource collection object.
        """
        return '<%s ResourceCollection at %s>' % (self._resource_name, self._url)


class Resource(object):
    """
    The Resource class represents a single REST API resource. All data is
    available as class members. The class members are automatically converted
    to their respective types using the resource schema. For instance,
    timestamps are converted into DateTime objects, integers are proper ints,
    and foreign keys point to other Resource objects.
    """

    def __init__(self, api, collection, id, data={}):
        self._api = api
        self._collection = collection
        self._id = id
        if self._id:
            self._url = productstatus.utils.build_url(self._collection._url, self._id)
        else:
            self._url = None
        self._data = copy.copy(data)
        self._unserialize()

    def save(self):
        """
        Store the locally cached values on the server.
        """
        if self._has_url():
            response = self._api._do_request('put', self._url, data=self._serialize())
        else:
            response = self._api._do_request('post', self._collection._url, data=self._serialize())
            self._url = productstatus.utils.build_url(self._api._base_url, response.headers['Location'])
        self._data = {}  # invalidate local cache

    def _has_url(self):
        """
        Returns True if this Resource has an URL which can be accessed at the
        server, False otherwise.
        """
        return self._url is not None

    def _get_resource_from_server(self):
        """
        Fetch the resource from the API server.
        """
        if not self._has_url():
            raise productstatus.exceptions.ProductstatusException('Trying to get an object without a primary key')
        try:
            response = self._api._do_request('get', self._url)
            self._data = self._api._get_response_data(response)
        except productstatus.exceptions.NotFoundException, e:
            raise productstatus.exceptions.ResourceNotFoundException(e)
        self._unserialize()

    def _ensure_complete_object(self):
        """
        Fetch the resource from the API server if we have an URL and it is not
        already cached.
        """
        if self._has_url() and not self._data:
            self._get_resource_from_server()

    def _serialize(self):
        """
        Return a JSON serialized representation of this resource.
        """
        self._ensure_complete_object()
        data = {}
        for key in self._data.keys():
            data[key] = self._serialize_member(key)
        return json.dumps(data)

    def _serialize_member(self, name):
        """
        Serialize a resource variable into a string, integer, boolean, or null.
        """
        if self._data[name] is None:
            return None

        description = self._collection.schema['fields'][name]
        type_ = description['type']
        if type_ == 'datetime':
            return self._data[name].strftime('%Y-%m-%dT%H:%M:%S%z')
        elif type_ == 'related' and description['related_type'] == 'to_one':
            return self._data[name].resource_uri
        return self._data[name]

    def _unserialize(self):
        """
        Replace all string members with their proper types.
        """
        for member in self._data.keys():
            self._unserialize_member(member)

    def _unserialize_member(self, name):
        """
        Convert string data into their proper types, according to the resource schema.
        """
        if self._data[name] is None:
            return

        description = self._collection.schema['fields'][name]
        type_ = description['type']
        if type_ == 'integer':
            self._data[name] = int(self._data[name])
        elif type_ == 'datetime':
            self._data[name] = dateutil.parser.parse(self._data[name])
        elif type_ == 'related' and description['related_type'] == 'to_one':
            self._data[name] = self._api[self._data[name]]

    def __getattr__(self, name):
        """
        Attribute accessor. Will load data from the server unless it is cached.
        Enables lazy loading of the resource.
        """
        fields = self._collection.schema['fields']
        if name not in fields:
            raise AttributeError('Attribute does not exist: %s' % name)
        self._ensure_complete_object()
        if name not in self._data:
            return None
        return self._data[name]

    def __setattr__(self, name, value):
        """
        Attribute setter. Will store data cached locally until saved using save().
        Will only allow setting variables that can be stored on the server, and
        disallow changing read-only attributes.
        """
        if name[0] == '_':
            return object.__setattr__(self, name, value)
        fields = self._collection.schema['fields']
        if name not in fields:
            raise KeyError('Attribute does not exist: %s' % name)
        if fields[name]['readonly']:
            raise AttributeError('Attribute is read only: %s' % name)
        # FIXME: more tests?
        self._ensure_complete_object()
        self._data[name] = value

    def __repr__(self):
        """
        Return a human-readable string representing this resource object.
        """
        if self._has_url():
            return '<Resource at %s>' % self.resource_uri
        return '<non-persistent %s Resource>' % self._collection._resource_name


class TastypieApiKeyAuth(requests.auth.AuthBase):
    """
    Django Tastypie requires a special Authorization header format, which is
    implemented by this class.
    """

    def __init__(self, username, api_key):
        self.username = username
        self.api_key = api_key

    def __call__(self, request):
        request.headers.update({'Authorization': 'ApiKey %s:%s' % (self.username, self.api_key)})
        return request
